/***********************************************************
 * $Id: UtcTimestamp.java 45 2007-06-29 13:19:57Z wglas $
 * 
 * Utility classes of the clazzes.org project
 * http://www.clazzes.org
 *
 * Created: 2006-03-13
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 * 
 ***********************************************************/

package org.clazzes.util.datetime;

import java.io.Serializable;
import java.text.ParseException;
import java.util.Calendar;
import java.util.TimeZone;

import javax.persistence.Embeddable;
import javax.persistence.Transient;

/**
 * A class to store UTC timestamps and timezone offset to database.
 * 
 * @author rbreuss
 * 
 */
@Embeddable
public class UtcTimestamp implements Serializable, Comparable<UtcTimestamp> {

    /**
     * To be changed at class-modifcation.
     */
    private static final long serialVersionUID = -4171006885461716579L;
    
    private static final long REF_TIMESTAMP = 2440588L *  24L * 3600000L;
    
    private static final int DAYS_IN_MONTH[] =
        new int [] { 29,31,28,31,30,31,30,31,31,30,31,30,31 };
    
    /**
     * This class is designed to handle UTC-Timestamps, so this field must not
     * be changed. It is used internally to calculate appropreate Calendar
     * Objects.
     */
    private static final TimeZone utcTimezone = TimeZone.getTimeZone("UTC");

    private static class DateFields {
        public final int year;
        public final int month;
        public final int day;
        
        DateFields(int year, int month, int day) {
            this.year = year;
            this.month = month;
            this.day = day;
        }
    };
    
    /**
     * The amount of milliseconds elapsed since 1970-01-01.
     */
    private Long utcMillis;

    /**
     * The timezone offset in minutes.
     */
    private Integer tzOffset;

    /**
     * A cache for decomposed year/month/day
     */
    private transient DateFields dateFields;
    
    @Transient
    private long getLocalTimestamp() {
        if(this.utcMillis==null) return 0L;
        
        long ts = this.utcMillis;
        
        if (this.tzOffset != null) ts += this.tzOffset * 60000L;
        return ts + REF_TIMESTAMP;
    }
    
    private void initDateFields() {
        
        if (this.utcMillis == null) {
            this.dateFields = new DateFields(0,0,0);
            return;
        }
        
        long linear_day = this.getLocalTimestamp();
        
        linear_day /= (24L * 3600000L);
        
        if (linear_day < 2363522L /* 1759-01-01 */ ||
                linear_day > 5373484L /* 9999-12-31 */   )
        {
            this.dateFields = new DateFields(0,0,0);
            return;
        }

        long x;
        long j = linear_day - 1721119L;
        long y = (j*4 - 1)/146097;
        j = j*4 - 146097*y - 1;
        x = j/4;
        j = (x*4 + 3) / 1461;
        y = 100*y + j;
        x = (x*4) + 3 - 1461*j;
        x = (x + 4)/4;
        long m = (5*x - 3)/153;
        x = 5*x - 3 - 153*m;
        long d = (x + 5)/5;

        if ( m < 10 ) {
            m += 3;
        } else {
            m -= 9;
            ++y;
        }
         
        this.dateFields = new DateFields((int)y,(int)m,(int)d);
    }
    
    private static long makeUtcMillis (int year, int month, int day) {
        
        long c, ya;
        long y = year;
        long m = month;
        if ( m > 2 ) {
          m -= 3;
        } else {
          m += 9;
          y--;
        }
        c = y;
        c /= 100;
        ya = y - 100*c;
        long linear_day = 1721119 + day + (146097*c)/4 + (1461*ya)/4 + (153*m+2)/5;
        linear_day -= 2440588L;
        return linear_day * 24L * 3600000L;
    }
    
    /**
     * @param y The year in the range from 1759 to 9999.
     * @return Whether the given year is a leap year in the gregorian calendar.
     */
    public static boolean isLeapYear (int y)
    {
      return (y % 4) == 0 && ((y % 400) == 0 || (y % 100) != 0); 
    }

    /**
     * @param y The year to consider.
     * @param m The month to consider.
     * @return The number of days in the given month.
     */
    public static int getDaysOfMonth (int y, int m)
    {
      if (m < 1 || m > 12) return 0;

      if (m == 2 && isLeapYear(y)) m = 0;

      return DAYS_IN_MONTH[m];
    }
    
    /**
     * Construct a new UtcTimestamp for 1970-01-01 (tzOffset=0).
     */
    public UtcTimestamp() {
        this(0L, 0);
    }
    /** Construct a new UtcTimestamp for the passed Calendar object.
     * @param c
     */
    public UtcTimestamp(Calendar c) {
        super();
        this.setFromCalendar(c);
    }
    /**
     * Construct a new UtcTimestamp for the passed milliseconds with a tzOffset=0.
     * @param millis
     */
    public UtcTimestamp(long millis) {
        this(millis, 0);
    }
    /**
     * Construct a new UtcTimestamp for the passed milliseconds an timezone offset.
     * @param millis
     * @param tzOffset in minutes
     */
    public UtcTimestamp(long millis, int tzOffset) {
        this.utcMillis = Long.valueOf(millis);
        this.tzOffset = Integer.valueOf(tzOffset);
    }
    
    /**
     * Contruct an UtcTimestamp from an ISO8601-formatted string.
     * 
     * @param s The string to parse.
     * @throws ParseException If the timestmap is not in the ISO8601 format.
     */
    public UtcTimestamp(String s) throws ParseException {
        this.setString(s);
    }

    /**
     * Convert this UtcTimestamp to java.util.Calendar.
     * @return
     */
    public Calendar toCalendar() {
        if (this.utcMillis == null || this.tzOffset == null)
        {
            return null;
        }
        else
        {
            Calendar c = Calendar.getInstance(UtcTimestamp.utcTimezone);
            c.setTimeInMillis(this.utcMillis.longValue() + (long) this.tzOffset.intValue() * 60000);
            c.set(Calendar.ZONE_OFFSET, this.tzOffset.intValue() * 60000);
            return c;
        }
    }
    
    /** 
     * Convert the passed Calendar object to UtcTimestamp.
     * @param c
     */
    @Transient
    public void setFromCalendar(Calendar c) {
        if (c != null) 
        {
            this.utcMillis = Long.valueOf(c.getTimeInMillis());
            this.tzOffset = Integer.valueOf((c.get(Calendar.ZONE_OFFSET) + c.get(Calendar.DST_OFFSET)) / 60000);
        }
        else
        {
            this.utcMillis = null;
            this.tzOffset = null;
        }
        this.dateFields = null;
    }
    
    /**
     * @return Returns the tzOffset.
     */
    public Integer getTzOffset() {
        return this.tzOffset;
    }

    /**
     * @param tzOffset
     *            The tzOffset to set.
     */
    public void setTzOffset(Integer tzOffset) {
        this.tzOffset = tzOffset;
        this.dateFields = null;
    }

    /**
     * @return Returns the utcMillis.
     */
    public Long getUtcMillis() {
        return this.utcMillis;
    }

    /**
     * @param utcMillis
     *            The utcMillis to set.
     */
    public void setUtcMillis(Long utcMillis) {
        this.utcMillis = utcMillis;
        this.dateFields = null;
    }
    
    /**
     * Set the day relative to the current timezone.
     * 
     * @param year The year in the range from 1759 to 9999.
     * @param month The month in the range from 1 to 12.
     * @param day The month in the range from 1 to 31.
     */
    @Transient
    public void setDate(int year, int month, int day) {
        
        long ts = makeUtcMillis(year, month, day);
        if (this.tzOffset != null)
            ts -= this.tzOffset * 60000L;
            
        this.utcMillis = ts;
        this.dateFields = null;
    }
    
    /**
     * @param hour
     * @param minute
     * @param second
     */
    @Transient
    public void setTimeOfDay(int hour, int minute, int second) {
        setTimeOfDay(hour, minute, second, 0);
    }
    
    /**
     * @param hour
     * @param minute
     * @param second
     * @param millis
     */
    @Transient
    public void setTimeOfDay(int hour, int minute, int second, int millis) {
        long ts = this.getLocalTimestamp();
        
        ts = (ts / (24L * 3600000L)) * (24L * 3600000L) +
        hour * 3600000L + minute * 60000L + second * 1000L + (long) millis;
        
        if (this.tzOffset != null)
            ts -= this.tzOffset * 60000L;
        
        this.utcMillis = ts - REF_TIMESTAMP;
        this.dateFields = null;
    }
    
    /**
     * @return The year in the range from 1759 to 9999.
     *         The returned month is relative to the local timezone, if a
     *         timezone offset is set. Zero is returned, if the thimestamp
     *         has been set to an illegal value.
     */
    @Transient
    public int getYear() {
        if (this.dateFields == null) initDateFields();
        return this.dateFields.year;
    }
    
   /**
    * @return The month in the range from 1 to 12.
    *         The returned month is relative to the local timezone, if a
    *         timezone offset is set.
    */
    @Transient
    public int getMonth() {
        if (this.dateFields == null) initDateFields();
        return this.dateFields.month;
    }
    
    /**
     * @return The day of month in the range from 1 to 31.
     *         The returned day is relative to the local timezone, if a
     *         timezone offset is set.
     */
    @Transient
    public int getDay() {
        if (this.dateFields == null) initDateFields();
        return this.dateFields.day;
    }
    
    /**
     * @return The day of the week represented by this timestamp in the
     *         timezone designated by tzOffset.
     *         
     * @see Calendar#SUNDAY
     * @see Calendar#MONDAY
     * @see Calendar#TUESDAY
     * @see Calendar#WEDNESDAY
     * @see Calendar#THURSDAY
     * @see Calendar#FRIDAY
     * @see Calendar#SATURDAY
     */
    @Transient
    public int getDayOfWeek() {
        return (int)((this.getLocalTimestamp() / (24L * 3600000L) + 1)%7L) + 1;
    }
    
    /**
     * @return The number of day in the month corresponding to the
     *         timestamp in the local timezone.
     */
    @Transient
    public int getDaysOfMonth()
    {
        if (this.dateFields == null) initDateFields();
        return getDaysOfMonth(this.dateFields.year,this.dateFields.month);
    }

    /**
     * @return Whther the year corresponding to the
     *         timestamp in the local timezone is a leap year.
     */
    @Transient
    public boolean isLeapYear() {
        if (this.dateFields == null) initDateFields();
        return isLeapYear(this.dateFields.year);
    }
    
    /**
     * @return The hour of the day in the range from 0 to 23.
     *         The returned hour is relative to the local timezone, if a
     *         timezone offset is set.
     */
    @Transient
    public int getHour() {
        return (int)((this.getLocalTimestamp()/3600000L)%24L);
    }
    
    /**
     * @return The minutes in the range from 0 to 59.
     */
    @Transient
    public int getMinute() {
        return (int)((this.getLocalTimestamp()/60000L)%60L);
    }
    
    /**
     * @return The seconds in the range from 0 to 59.
     */
    @Transient
    public int getSecond() {
        return (int)((this.getLocalTimestamp()/1000L)%60L);
    }
    
    /**
     * @return The milliseconds in the ragne from 0 to 999.
     */
    @Transient
    public int getMilliSecond() {
        return (int)(this.getLocalTimestamp()%1000L);
    }
    
    /**
     * @param ms The number of milliseconds to add.
     */
    public void addMilliSeconds(long ms) {
        this.utcMillis += ms;
        this.dateFields = null;
    }
    
    /**
     * @param s The number of seconds to add.
     */
    public void addSeconds(long s) {
        this.addMilliSeconds(s * 1000L);
    }
    
    /**
     * @param m The number of minutes to add.
     */
    public void addMinutes(long m) {
        this.addMilliSeconds(m * 60000L);
    }
    
    /**
     * @param h The number of hours to add.
     */
    public void addHours(long h) {
        this.addMilliSeconds(h * 3600000L);
    }
    
    /**
     * @param d The number of days to add.
     */
    public void addDays(long d) {
        this.addMilliSeconds(d * 24L * 3600000L);
    }
    
    /**
     * Add <code>m</code> month to the given timestamp.
     * This method keeps the return timestamp on the last day in month, so
     * <code>UtcTimestamp("2007-02-28").addMonths(3) == UtcTimestamp("2007-05-31")</code> 
     *
     * @param m The number of month to add.
     */
    public void addMonths(int m) {
        if (this.dateFields == null) this.initDateFields();
        
        int mmm = this.dateFields.year * 12 + (this.dateFields.month-1) + m;
        
        int mm = mmm % 12 + 1;
        int yy = mmm / 12;
        
        int dd = this.dateFields.day;
        int dom = getDaysOfMonth(yy,mm);
        
        if (dd == getDaysOfMonth(this.dateFields.year, this.dateFields.month) ||
                dd > dom)
            dd = dom;
        
        long ts = makeUtcMillis(yy, mm, dd);
        ts += this.getLocalTimestamp() % (24L * 3600000L);
        
        if (this.tzOffset != null)
            ts -= this.tzOffset * 60000L;
            
        this.utcMillis = ts;
        this.dateFields = null;
    }
    
    /**
     * Add <code>m</code> month to the given timestamp.
     * This method keeps the return timestamp on the last day in month, so
     * <code>UtcTimestamp("2007-02-28").addYear(-7) == UtcTimestamp("2000-02-29")</code>. 

     * @param y The number of years to add.
     */
    public void addYears(int y) {
        if (this.dateFields == null) this.initDateFields();
         
        int dd = this.dateFields.day;
        
        if (this.dateFields.month == 2 &&
                dd == getDaysOfMonth(this.dateFields.year, this.dateFields.month))
            dd = getDaysOfMonth(this.dateFields.year+y, this.dateFields.month);
            
        long ts = makeUtcMillis(this.dateFields.year+y, this.dateFields.month, dd);
        ts += this.getLocalTimestamp() % (24L * 3600000L);
        
        if (this.tzOffset != null)
            ts -= this.tzOffset * 60000L;
            
        this.utcMillis = ts;
        this.dateFields = null;
    }
    
    /**
     * This method duplicated the funtionality of <code>Calendar.add()</code> in order
     * to make the transition from Calendar database obejcts to UtcTimestamp object
     * as easy as possible.
     * 
     * @param field The field to increment, one of {@link Calendar#YEAR},
     *           {@link Calendar#MONTH}, {@link Calendar#DAY}, {@link Calendar#DAY_OF_MONTH},
     *           {@link Calendar#HOUR}, {@link Calendar#HOUR_OF_DAY}, {@link Calendar#MINUTE},
     *           {@link Calendar#SECOND} or {@link Calendar#MILLISECOND}.
     * @param amount The amount to increment the given field.
     * 
     * @see Calendar#add(int, int)
     */
    public void add(int field, int amount) {
        switch(field) {
        case Calendar.YEAR:
            this.addYears(amount);
            break;
            
        case Calendar.MONTH:
            this.addMonths(amount);
            break;
            
        case Calendar.DAY_OF_MONTH:
            this.addDays(amount);
            break;
            
        case Calendar.HOUR:
        case Calendar.HOUR_OF_DAY:
            this.addHours(amount);
            break;

        case Calendar.MINUTE:
            this.addMinutes(amount);
            break;

        case Calendar.SECOND:
            this.addSeconds(amount);
            break;
            
        case Calendar.MILLISECOND:
            this.addMilliSeconds(amount);
            break;
        }
    }
    
    /**
     * @param format One of {@link ISO8601Format#DATE_FORMAT},
     *               {@link ISO8601Format#DATETIME_FORMAT} or {@link ISO8601Format#MILLISECOND_FORMAT}.
     * @return The formatted timestamp cf. to ISO8601.
     */
    public String toString (int format)
    {
        if (this.utcMillis == null) return null;
        
        if (this.dateFields == null) initDateFields();
        
        if (format == ISO8601Format.DATE_FORMAT) {
            
            return String.format("%04d-%02d-%02d",
                    this.dateFields.year,
                    this.dateFields.month,
                    this.dateFields.day);

        } else if (format == ISO8601Format.DATETIME_FORMAT) {
            long ts = this.getLocalTimestamp();
            
            if (this.tzOffset != null) {
            
                if (this.tzOffset == 0)
                    return String.format("%04d-%02d-%02dT%02d:%02d:%02dZ",
                            this.dateFields.year,
                            this.dateFields.month,
                            this.dateFields.day,
                            (int)((ts/3600000L)%24L),
                            (int)((ts/60000L)%60L),
                            (int)((ts/1000L)%60L) );
                else if (this.tzOffset > 0)
                    return String.format("%04d-%02d-%02dT%02d:%02d:%02d+%02d:%02d",
                            this.dateFields.year,
                            this.dateFields.month,
                            this.dateFields.day,
                            (int)((ts/3600000L)%24L),
                            (int)((ts/60000L)%60L),
                            (int)((ts/1000L)%60L),
                            this.tzOffset/60,
                            this.tzOffset%60);
                else
                    return String.format("%04d-%02d-%02dT%02d:%02d:%02d-%02d:%02d",
                            this.dateFields.year,
                            this.dateFields.month,
                            this.dateFields.day,
                            (int)((ts/3600000L)%24L),
                            (int)((ts/60000L)%60L),
                            (int)((ts/1000L)%60L),
                            (-this.tzOffset)/60,
                            (-this.tzOffset)%60);
                
            } else {
                
                return String.format("%04d-%02d-%02dT%02d:%02d:%02d",
                        this.dateFields.year,
                        this.dateFields.month,
                        this.dateFields.day,
                        (int)((ts/3600000L)%24L),
                        (int)((ts/60000L)%60L),
                        (int)((ts/1000L)%60L) );
            }
            
        }
        /* ISO8601Format.MILLISECOND_FORMAT */
        else {
            
            long ts = this.getLocalTimestamp();
        
            if (this.tzOffset != null) {
            
                if (this.tzOffset == 0)
                    return String.format("%04d-%02d-%02dT%02d:%02d:%02d.%03dZ",
                            this.dateFields.year,
                            this.dateFields.month,
                            this.dateFields.day,
                            (int)((ts/3600000L)%24L),
                            (int)((ts/60000L)%60L),
                            (int)((ts/1000L)%60L),
                            (int)(ts%1000L));
                else if (this.tzOffset > 0)
                    return String.format("%04d-%02d-%02dT%02d:%02d:%02d.%03d+%02d:%02d",
                            this.dateFields.year,
                            this.dateFields.month,
                            this.dateFields.day,
                            (int)((ts/3600000L)%24L),
                            (int)((ts/60000L)%60L),
                            (int)((ts/1000L)%60L),
                            (int)(ts%1000L),
                            this.tzOffset/60,
                            this.tzOffset%60);
                else
                    return String.format("%04d-%02d-%02dT%02d:%02d:%02d.%03d-%02d:%02d",
                            this.dateFields.year,
                            this.dateFields.month,
                            this.dateFields.day,
                            (int)((ts/3600000L)%24L),
                            (int)((ts/60000L)%60L),
                            (int)((ts/1000L)%60L),
                            (int)(ts%1000L),
                            (-this.tzOffset)/60,
                            (-this.tzOffset)%60);
                
            } else {
                
                return String.format("%04d-%02d-%02dT%02d:%02d:%02d.%03d",
                        this.dateFields.year,
                        this.dateFields.month,
                        this.dateFields.day,
                        (int)((ts/3600000L)%24L),
                        (int)((ts/60000L)%60L),
                        (int)((ts/1000L)%60L),
                        (int)(ts%1000L) );
            }
        }   
    }
    
    /**
     * Set an ISO8601 formatted timestamp to this object.
     * 
     * @throws ParseException If the given string is not formatted like ISO8601.
     */
    @Transient
    public void setString(String s) throws ParseException
    {
        int pos = 0;
        
        try {
            int y = Integer.valueOf(s.substring(pos,pos+4));
            pos+=4;
            if (s.charAt(pos) != '-') throw new ParseException(s,pos);
            ++pos;
            int m = Integer.valueOf(s.substring(pos,pos+2));
            pos+=2;
            if (s.charAt(pos) != '-') throw new ParseException(s,pos);
            ++pos;
            int d = Integer.valueOf(s.substring(pos,pos+2));
            pos += 2;
            
            if (pos < s.length()) {
                if (s.charAt(pos) != 'T') throw new ParseException(s,pos);
                ++pos;
                
                int hh = Integer.valueOf(s.substring(pos,pos+2));
                pos+=2;
                if (s.charAt(pos) != ':') throw new ParseException(s,pos);
                ++pos;
                int mm = Integer.valueOf(s.substring(pos,pos+2));
                pos+=2;
                if (s.charAt(pos) != ':') throw new ParseException(s,pos);
                ++pos;
                int ss = Integer.valueOf(s.substring(pos,pos+2));
                pos += 2;
               
                long ms = ((long)hh * 3600000L) + ((long)mm * 60000L) + ((long)ss * 1000L);
                
                if (pos < s.length() && s.charAt(pos) == '.') {
                    ++pos;
                    int l = 0;
                    while (pos+l < s.length() && Character.isDigit(s.charAt(pos+l))) ++l;
                
                    int sss= Integer.valueOf(s.substring(pos,pos+(l > 3 ? 3 : l)));
                    switch (l) {
                    case 1: sss *= 100; break;
                    case 2: sss *= 10; break;
                    }
                    
                    ms += sss;
                    pos+=l;
                 }
 
                Integer tz_off = null;
                
                if (pos < s.length()) {
                    if (s.charAt(pos) == 'Z') {
                        ++pos;
                        tz_off = 0;
                    }
                    else if (s.charAt(pos) == '+') {
                        ++pos;
                        int tz_hh = Integer.valueOf(s.substring(pos,pos+2));
                        pos+=2;
                        if (s.charAt(pos) != ':') throw new ParseException(s,pos);
                        ++pos;
                        int tz_mm = Integer.valueOf(s.substring(pos,pos+2));
                        pos+=2;
                        
                        tz_off = tz_hh*60 + tz_mm;
                    }
                    else if (s.charAt(pos) == '-') {
                        ++pos;
                        int tz_hh = Integer.valueOf(s.substring(pos,pos+2));
                        pos+=2;
                        if (s.charAt(pos) != ':') throw new ParseException(s,pos);
                        ++pos;
                        int tz_mm = Integer.valueOf(s.substring(pos,pos+2));
                        pos+=2;
                        
                        tz_off = tz_hh*-60 - tz_mm;
                    }
                    else 
                        throw new ParseException(s,pos);
                    
                    if (pos<s.length())
                        throw new ParseException(s,pos);
                }
                
                
                this.setTzOffset(tz_off);
                this.setDate(y,m,d);
                this.addMilliSeconds(ms);
            }
            else {
                this.setTzOffset(null);
                this.setDate(y,m,d);
            }
            
        } catch(NumberFormatException e) {
            throw new ParseException(s,pos);
        }
        
    }
    
    
    /**
     * @return The result of <code>toString(ISO8601Format.MILLISECOND_FORMAT)</code>.
     * 
     * @see #toString(int)
     * @see ISO8601Format#DATE_FORMAT
     * @see ISO8601Format#DATETIME_FORMAT
     * @see ISO8601Format#MILLISECOND_FORMAT
     */
    public String toString() {
        
        return this.toString(ISO8601Format.MILLISECOND_FORMAT);
    }
    
    /* (non-Javadoc)
     * @see java.lang.Comparable#compareTo(java.lang.Object)
     */
    public int compareTo(UtcTimestamp o) {
        if (o == null) 
            return 1;
        if (this.utcMillis == null)
        {
            if (o.utcMillis == null)
                return 0;
            else
                return -1;
        }
        else
        {
            if (o.utcMillis == null)
                return 1;
            else
                return ((this.utcMillis < o.utcMillis) ? -1
                        : (this.utcMillis > o.utcMillis) ? 1
                        : 0);
        }
    }

    /* (non-Javadoc)
     * @see java.lang.Object#equals(java.lang.Object)
     */
    public boolean equals(Object o)
    {
        if (o == null || !(o instanceof UtcTimestamp)) return false;
        
        UtcTimestamp ts = (UtcTimestamp)o;
        
        if (this.getUtcMillis() == null) {
            return ts.getUtcMillis() == null;
        }
        
        return this.getUtcMillis().equals(ts.getUtcMillis());
    }
    
    /* (non-Javadoc)
     * @see java.lang.Object#hashCode()
     */
    public int hashCode() {
        if (this.getUtcMillis() == null) 
            return -1;
        
        return (int) (419430343L /* prime */ * this.getUtcMillis());
    }
}
