/***********************************************************
 * $Id: PropertyHelper.java 4 2006-12-26 19:16:57Z wglas $
 * 
 * Utility classes of the clazzes.org project
 * http://www.clazzes.org
 *
 * Created: 2006-03-13
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 * 
 ***********************************************************/

package org.clazzes.util.reflect;


/**
 * This class provides some static function for transformting
 * property names to getter/setter names.
 * 
 * @author wglas
 */
public class PropertyHelper
{
    /**
     * Set to private, because we only provide for static functions.
     */
    private PropertyHelper()
    {}
    
    /**
     * @param property The property for which to retrieve the getter name.
     * @return The name of the getter method for the given property.
     * 
     * @see PropertyHelper#getBooleanGetterName(String)
     */
    public static String getGetterName(String property)
    {
        return "get" + Character.toUpperCase(property.charAt(0)) + property.substring(1);
    }
    
    /**
     * @param property The property for which to retrieve the setter name.
     * @return The name of the setter method for the given property.
     */
    public static String getSetterName(String property)
    {
        return "set" + Character.toUpperCase(property.charAt(0)) + property.substring(1);
    }
    
    /**
     * This method complements the functionality of {@link PropertyHelper#getGetterName(String)}
     * for boolen properties, where the getter method start with the string "is" instead of "get".
     * 
     * @param property The boolean property for which to retrieve the setter name.
     * @return The name of the setter method for the given property.
     */
    public static String getBooleanGetterName(String property)
    {
        return "is" + Character.toUpperCase(property.charAt(0)) + property.substring(1);
    }
    
    private static String extractProperty(String pfx, String methodName)
    {
        if (methodName == null ||
                methodName.length() < pfx.length()+1 ||
                !methodName.startsWith(pfx) ||
                !Character.isUpperCase(methodName.charAt(pfx.length())))
            return null;

        if (methodName.length() < pfx.length()+2 ||
                Character.isLowerCase(methodName.charAt(pfx.length()+1)))
        {
            // getXy -> property is xy
            return
            Character.toLowerCase(methodName.charAt(pfx.length())) +
            methodName.substring(pfx.length()+1);
        }
        else
        {
            // getXY -> property is XY.
            return methodName.substring(pfx.length());
        }
        
    }
    
    /**
     * Returns the property name of a given getter method name.
     * 
     * @param getterName The name of the getter function.
     * @return The associated property name or null, if the function name is not
     *         a valid getter name.
     *         
     * @see PropertyHelper#getBooleanPropertyFromGetterName(String)
     */
    public static String getPropertyFromGetterName(String getterName)
    {
       return extractProperty("get",getterName);
    }
    
    /**
     * Returns the property name of a given getter method name for boolean fields.
     * 
     * This method complements the functionality of {@link PropertyHelper#getGetterName(String)}
     * for boolen properties, where the getter method start with the string "is" instead of "get".
     *
     * @param getterName The name of the getter function.
     * @return The associated property name or null, if the function name is not
     *         a valid getter name.
     */
    public static String getBooleanPropertyFromGetterName(String getterName)
    {
       return extractProperty("is",getterName);
    }
    
    /**
     * Returns the property name of a given setter method name.
     * 
     * @param getterName The name of the setter function.
     * @return The associated property name or null, if the function name is not
     *         a valid setter name.
     */
    public static String getPropertyFromSetterName(String setterName)
    {
       return extractProperty("set",setterName);
    }
}
